from __future__ import print_function, unicode_literals, absolute_import, division
from six.moves import range, zip, map, reduce, filter

import math
import tensorflow as tf
from napari_hydra.hydrastardist.utils import _raise, backend_channels_last
from napari_hydra.hydrastardist.models.base_hydra import masked_loss_wbr
from csbdeep.utils.tf import keras_import
K = keras_import('backend')
Layer, Conv2D, MaxPooling2D, UpSampling2D, Conv3D, MaxPooling3D, UpSampling3D, Cropping2D, Cropping3D, Concatenate, Add, Dropout, Activation, BatchNormalization = \
    keras_import('layers', 'Layer', 'Conv2D', 'MaxPooling2D', 'UpSampling2D', 'Conv3D', 'MaxPooling3D', 'UpSampling3D', 'Cropping2D', 'Cropping3D', 'Concatenate', 'Add', 'Dropout', 'Activation', 'BatchNormalization')


# Layer that applies the WBR Loss
class ActivityRegularizationLayer(Layer):
    def __init__(self):
        super().__init__()

    def call(self, concat,  mask=None):
        self.add_loss(masked_loss_wbr(mask)(mask, concat))
        return concat, mask

def conv_block2(n_filter, n1, n2,
                activation="relu",
                border_mode="same",
                dropout=0.0,
                batch_norm=False,
                init="glorot_uniform",
                **kwargs):

    def _func(lay):
        if batch_norm:
            s = Conv2D(n_filter, (n1, n2), padding=border_mode, kernel_initializer=init, **kwargs)(lay)
            s = BatchNormalization()(s)
            s = Activation(activation)(s)
        else:
            s = Conv2D(n_filter, (n1, n2), padding=border_mode, kernel_initializer=init, activation=activation, **kwargs)(lay)
        if dropout is not None and dropout > 0:
            s = Dropout(dropout)(s)
        return s

    return _func



def conv_block3(n_filter, n1, n2, n3,
                activation="relu",
                border_mode="same",
                dropout=0.0,
                batch_norm=False,
                init="glorot_uniform",
                **kwargs):

    def _func(lay):
        if batch_norm:
            s = Conv3D(n_filter, (n1, n2, n3), padding=border_mode, kernel_initializer=init, **kwargs)(lay)
            s = BatchNormalization()(s)
            s = Activation(activation)(s)
        else:
            s = Conv3D(n_filter, (n1, n2, n3), padding=border_mode, kernel_initializer=init, activation=activation, **kwargs)(lay)
        if dropout is not None and dropout > 0:
            s = Dropout(dropout)(s)
        return s

    return _func


def unet_down_block(n_depth=2, n_filter_base=16, kernel_size=(3,3), n_conv_per_depth=2,
               activation="relu",
               batch_norm=False,
               dropout=0.0,
               last_activation=None,
               pool=(2,2),
               kernel_init="glorot_uniform",
               expansion=2,
               prefix=''):

    if len(pool) != len(kernel_size):
        raise ValueError('kernel and pool sizes must match.')
    n_dim = len(kernel_size)
    if n_dim not in (2,3):
        raise ValueError('unet_block only 2d or 3d.')

    conv_block = conv_block2  if n_dim == 2 else conv_block3
    pooling    = MaxPooling2D if n_dim == 2 else MaxPooling3D
    upsampling = UpSampling2D if n_dim == 2 else UpSampling3D

    if last_activation is None:
        last_activation = activation

    channel_axis = -1 if backend_channels_last() else 1

    def _name(s):
        return prefix+s

    def _func(input):
        skip_layers = []
        layer = input

        # down ...
        for n in range(n_depth):
            for i in range(n_conv_per_depth):
                layer = conv_block(int(n_filter_base * expansion ** n), *kernel_size,
                                   dropout=dropout,
                                   activation=activation,
                                   init=kernel_init,
                                   batch_norm=batch_norm, name=_name("down_level_%s_no_%s" % (n, i)))(layer)
            skip_layers.append(layer)
            layer = pooling(pool, name=_name("max_%s" % n))(layer)

        # middle
        for i in range(n_conv_per_depth - 1):
            layer = conv_block(int(n_filter_base * expansion ** n_depth), *kernel_size,
                               dropout=dropout,
                               init=kernel_init,
                               activation=activation,
                               batch_norm=batch_norm, name=_name("middle_%s" % i))(layer)

        layer = conv_block(int(n_filter_base * expansion ** max(0, n_depth - 1)), *kernel_size,
                           dropout=dropout,
                           activation=activation,
                           init=kernel_init,
                           batch_norm=batch_norm, name=_name("middle_%s" % n_conv_per_depth))(layer)


        return layer,  skip_layers

    return _func

def unet_up_block(n_depth=2,
                  n_filter_base=16,
                  kernel_size=(3,3),
                  n_conv_per_depth=2,
                  activation="relu",
                  batch_norm=False,
                  dropout=0.0,
                  last_activation=None,
                  pool=(2,2),
                  kernel_init="glorot_uniform",
                  expansion=2,
                  prefix='',
                  branch='1'):

    if len(pool) != len(kernel_size):
        raise ValueError('kernel and pool sizes must match.')
    n_dim = len(kernel_size)
    if n_dim not in (2,3):
        raise ValueError('unet_block only 2d or 3d.')

    conv_block = conv_block2  if n_dim == 2 else conv_block3
    pooling    = MaxPooling2D if n_dim == 2 else MaxPooling3D
    upsampling = UpSampling2D if n_dim == 2 else UpSampling3D

    channel_axis = -1 if backend_channels_last() else 1

    def _name(s):
        return prefix+s

    def _func(layer,skip_layers):

        # ...and up with skip layers
        # Need to make two copies of this up layers
        for n in reversed(range(n_depth)):
            layer = Concatenate(axis=channel_axis,name=_name("concat_%s_%s" % (branch,n)))([upsampling(pool)(layer), skip_layers[n]])
            for i in range(n_conv_per_depth - 1):
                layer = conv_block(int(n_filter_base * expansion ** n), *kernel_size,
                                   dropout=dropout,
                                   init=kernel_init,
                                   activation=activation,
                                   batch_norm=batch_norm, name=_name("up_level_%s_no%s_%s" % (n, branch,i)))(layer)

            layer = conv_block(int(n_filter_base * expansion ** max(0, n - 1)), *kernel_size,
                               dropout=dropout,
                               init=kernel_init,
                               activation=activation if n > 0 else last_activation,
                               batch_norm=batch_norm, name=_name("up_level_%s_no%s_%s" % (n, branch,n_conv_per_depth)))(layer)
        return layer

    return _func


import numpy as np
from napari_hydra.utils.rasterize_labels import rasterize_labels

def test_rasterize_labels():
    # Define a simple square polygon
    # Coords: (2,2), (2,5), (5,5), (5,2)
    # Label dict format: 'coord': [[X_arr, Y_arr]]
    coord_x = np.array([2, 2, 5, 5])
    coord_y = np.array([2, 5, 5, 2])
    
    label_dict = {"coord": [[coord_x, coord_y]]}
    out_shape = (10, 10)
    
    labels, polys = rasterize_labels(label_dict, out_shape)
    
    assert labels.shape == (10, 10)
    assert np.any(labels == 1)
    # Center should be filled
    assert labels[3, 3] == 1
    assert labels[0, 0] == 0



import json
from pathlib import Path
from typing import Union

from albumentations.core import serialization
from albumentations.core.serialization import Serializable as Transform


def read_json_config(config_file_path: Union[str, Path]) -> dict:
    with open(config_file_path, "r", encoding="UTF-8") as config_file:
        config = json.load(config_file)
    return config

def transform_to_json_string(transform: Transform) -> str:
    """ Takes an albumentations transform, serializes it and outputs a string representing it.

    Args:
        transform (Transform): Transform to serialize.

    Returns:
        str: String representing serialised transform.
    """

    transform_dict = serialization.to_dict(transform)
    return json.dumps(transform_dict, indent=4)


def transform_to_json(transform: Transform, json_file_path: Union[Path, str]) -> None:
    """ Takes an albumentations transform, serializes it and saves it to a json file.

    Args:
        transform (Transform): Transform to serialize.
        json_file_path (Union[Path, str]): path to a file in which to save transform config.

    """
    json_file_path = Path(json_file_path)
    if not json_file_path.parent.exists():
        json_file_path.parent.mkdir(parents=True)

    transform_dict = serialization.to_dict(transform)
    with open(json_file_path, "w", encoding="UTF-8") as json_file:
        json.dump(transform_dict, json_file, indent=4)
